require 'base.scripts.vec'

-- specify all these hud render coords in some standard aspect ratio like 640x480 or so
screenSize = vec2:new(640,480)

-- TODO - explicitly ask for the console font -- and expose a getter for the font character widths
fontTex = spriteLoad('../res/font.png')

function drawScreenText(x,y,w,h,str)
	local camera = viewGetBox()
	drawText{
		pos=vec2:new(x,y):divTo(screenSize):map(camera),
		size=vec2:new(w,h):divTo(screenSize):mulTo(camera:size()),
		str=str
	}
end

Text = {}
Text.color = vec4:new(1)
Text.letterSize = vec2:new(.5)
Text.depth = -2		-- with all these defaults you'd think i'd just class/inst them

function drawText(args)	--_pos,size,str)
	local pos = vec2:new(args.pos)
	local str = args.str
	local size = args.size or Text.letterSize -- hmm circular def'ing ... n.g.
	local color = args.color or Text.color
	local depth = args.depth or Text.depth
	local blend = args.blend or BLEND.NONE
	
	local startX = pos[1]
	
	for i=1,#str do
		local chbyte = str:byte(i)
		if chbyte == 10 or chbyte == 13 then
			pos[2] = pos[2] + size[2]
			pos[1] = startX
		else
			local ch = chbyte - 32
			
			local start, finish = getFontCharWidth(chbyte)
			local charWidth
			if start then
				charWidth = finish - start
			else
				start = 0
				finish = 1
				charWidth = 1
			end
			local tx = (math.floor(math.fmod(ch, 16)) + start) / 16
			local ty = math.floor(ch / 16) / 16
			local tsx = charWidth/16
			local tsy = 1/16
			local vtx = {
				pos[1], pos[2],
				pos[1] + size[1] * charWidth, pos[2],
				pos[1] + size[1] * charWidth, pos[2] + size[2],
				pos[1], pos[2] + size[2],
			}
			local tc = {
				tx, ty,
				tx+tsx, ty,
				tx+tsx, ty+tsy,
				tx, ty+tsy
			}
			drawGeometry(
			--geometry info
				vtx,
				tc,
				depth,			--depth
			--render info
				fontTex,
				color,		--color
				0,			--shader
				blend,
				args.alphaTest,		--no alpha
				args.dot3,		--no dot3
				args.lighting,		--no lighting
				args.postFBO or true,			-- draw outside (on top) of the fbo stuff
			--bumpmap tangent-space adjustment
				1,
				0,
				1);
			pos[1] = pos[1] + size[1] * charWidth
			if args.width then
				if pos[1] + size[1] - startX > args.width then
					pos[2] = pos[2] + size[2]
					pos[1] = startX
				end
			end
		end
	end
end
