require 'base.scripts.ents'
require 'data.scripts.weapons'	-- only for the weapons enum -- should this be a separate file?  a defs file?

entclass('Item', Collidable, {
	bbox = box2:new(.45);
	ctor = function(self,...)
		Item.super.ctor(self,...)
		self:setTouchOnly(true)
	end,
})

animLoad('health', {sprites={{file='items/iconh_red.tga',bbox={-1,-1,1,1}}}})

entclass('HealthItem', Item)

HealthItem.animGroup = 'health'
HealthItem.health = 25

function HealthItem:ctor(so)
	HealthItem.super.ctor(self, so)
	HealthItem:readSpawnParams(self, so)
end

function HealthItem:touch(other)
	HealthItem.super.touch(self,other)
	if not other or not other:isa(Player) or other.health <= 0 then return end
	other.health = other.health + self.health
	other.screenFlash:set(1)
	other.screenFlashEndTime = getTime() + 0.05
	self:delete()
end

SetupSpawnParams(HealthItem, {
	health = 'how much to heal';
})

Spawnable(HealthItem)


ITEM = enum({'JETPACK', 'FLIPOUT'}, 0)

animLoad('jetpack', {sprites={{file='items/jetpack1.tga',bbox={-.5,-1,.5,1}}}})

entclass('JetpackItem', Item, {
	animGroup = 'jetpack';
	duration = 30,
	flag = 1,
	touch = function(self,other)
		JetpackItem.super.touch(self,other)
		if not other or not other:isa(Player) or other.health <= 0 then return end
		other.itemFlags = bitor(other.itemFlags, self.flag)
		if other.jetpackEndTime < getTime() then
			other.jetpackEndTime = getTime()
		end
		other.jetpackEndTime = other.jetpackEndTime + self.duration
		self:delete()
	end,
})
Spawnable(JetpackItem)

animLoad('flipout', {sprites={{file='items/cancel.tga',bbox={-1,-1,1,1}}}})

entclass('FlipoutItem', Item, {
	animGroup = 'flipout';
	duration = 30,
	flag = 2,
	touch = function(self,other)
		FlipoutItem.super.touch(self,other)
		if not other or not other:isa(Player) or other.health <= 0 then return end
		other.itemFlags = bitor(other.itemFlags, self.flag)
		if other.flipoutEndTime < getTime() then
			other.flipoutEndTime = getTime()
		end
		other.flipoutEndTime = other.flipoutEndTime + self.duration
		self:delete()
	end,
})
Spawnable(FlipoutItem)

entclass('WeaponItem', Item, {
	ctor = function(self,so,weapon)
		WeaponItem.super.ctor(self,so)
		self.weapon = weapon
	end,
	touch = function(self,other)
		WeaponItem.super.touch(self,other)
		if not other or not other:isa(Player) or other.health <= 0 then return end
		-- subtract 1 so we can save 1 as the first lua enum (keeping things on par with Lua)
		-- while sending to C 0 as the first enum (keeping things on par with C)
		other.weaponFlags = bitor(other.weaponFlags, bitshiftleft(1, self.weapon))
		other.weapon = self.weapon
		self:delete()
	end
})

for i,w in ipairs(weapons) do
	Spawnable(entclass(w.spawnName, WeaponItem, {animGroup = w.groupName, weapon = i}))
end
