require 'base.scripts.ents'

entclass('JumpEnemy', TakesDamage, {

	STATE = {
		STAND = {
			think = function(self)
				self:setAnimSeq('idle')
				self:findTarget()
			end,
		},
		RUN = {
			think = function(self)
				self:setAnimSeq('run')
				self:run()
			end,
		},
		JUMP = {
			think = function(self)
				self:setAnimSeq('jump')
				self:run()
			end,
		},
		DIE = {},
	}
})

-- break so we can ref STATE (is this needed?)

union(JumpEnemy, {

	-- if you do state transition functions (like the Bat in wellstory) then this needs to be in the ctor
	state = JumpEnemy.STATE.STAND,

	jumpVel = 25,
	runMaxVel = 20,
	runForce = 15,
	airForce = 15,

	think = function(self)
		JumpEnemy.super.think(self)
		-- needed for any whose parent think() might kill themselves (Collidable)
		if not self:isActive() then return end	

		if self.state and self.state.think then self.state.think(self) end
	end,

	doJump = function(self)
		if self:getContactNormalY() < -0.5 then
			self:setVelY(self:getVelY() - self.jumpVel)
			self.state = self.STATE.JUMP
		end
	end,

	findTarget = function(self)
		self.state = self.STATE.STAND
		for other in allents() do
			if other and other:isa(Player) and other.health > 0 then
				self.enemy = other
				self.state = self.STATE.RUN
				return true
			end
		end
		return false
	end,
		
	run = function(self)
		if not self:findTarget() then return end
		
		local onGround = self:getContactNormalY() < -0.5
		
		if self.enemy.health <= 0 then
			self.enemy = nil
		end
		if not self.enemy then
			self.state = self.STATE.STAND
			return
		end
		
		self:setLeft(self.enemy:getPosX() < self:getPosX())
		
		if self.enemy:getPosX() < self:getPosX() - 0.5 then
			if onGround then
				if self:getVelX() > -self.runMaxVel then
					self:setVelX(self:getVelX() - self.runForce * getDeltaTime())
					if self:getVelX() < -self.runMaxVel then
						self:setVelX(-self.runMaxVel)
					end
				end
			else
				if self:getVelX() > -self.runMaxVel then
					self:setVelX(self:getVelX() - self.airForce * getDeltaTime())
				end
			end
		elseif self.enemy:getPosX() > self:getPosX() + 0.5 then
			if onGround then
				if self:getVelX() < self.runMaxVel then
					self:setVelX(self:getVelX() + self.runForce * getDeltaTime())
					if self:getVelX() > self.runMaxVel then
						self:setVelX(self.runMaxVel)
					end
				end
			else
				if self:getVelX() < self.runMaxVel then
					self:setVelX(self:getVelX() + self.airForce * getDeltaTime())
				end
			end
		end
		
		if onGround and math.random(16) == 16 then
			self:doJump()
			return
		end
		
		if onGround then
			self.state = self.STATE.RUN
		else
			self.state = self.STATE.JUMP
		end
	end,
})

