require 'base.scripts.ents'
require 'data.scripts.weapons'
require 'data.scripts.items'
require 'base.scripts.lighttex'
require 'base.scripts.clientinput'
require 'base.scripts.client'		--createClientStartClass
require 'base.scripts.defs'
require 'data.scripts.hud'
require 'base.scripts.shader'
require 'base.scripts.gameapi'

	-- jetpack
animLoad('jetpackflame', {
	sprites = {
		{file='weapons/muzzleflash copy.tga', bbox={-1,-.5,1,.5}};
		{file='weapons/muzzleflash copy.tga', bbox={-.75,-.375,.75,.375}};
		{file='weapons/muzzleflash copy.tga', bbox={-.5,-.25,.5,.25}};
	};
	seqs = {
		idle = {
			{joint = 1; time = 0; sprite = 1};
			{joint = 1; time = .1; sprite = 2};
			{joint = 1; time = .2; sprite = 3};
			{joint = 1; time = .3; sprite = 3};
		};
	};
})

entclass('JetpackFlame', Renderable, {
	ctor = function(self,owner)
		JetpackFlame.super.ctor(self)
		self:setLeft(owner:getLeft())
		self:setPosY(owner:getPosY() - 1)
		if owner:getLeft() then self:setPosX(owner:getPosX() + 1.5)
		else self:setPosX(owner:getPosX() - 1.5)
		end
		self.parent = owner
		self:setBlend(BLEND.ADD)
		self:setAnimGroup('jetpackflame')
		-- animEndFunc = RemoveEntity ...
		setTimeout(function() self:delete() end, 0.1)
	end,
	
	think = function(self)
		JetpackFlame.super.think(self)
		local o = self.parent
		if not o then return end
		self:setPosX(o:getPosX())
		self:setPosY(o:getPosY() - 1)
		
		if o:isa(Renderable) then
			if o:getLeft() then self:setPosX(o:getPosX() + 1.5)
			else self:setPosX(o:getPosX() - 1.5)
			end
			self:setLeft(o:getLeft())
		end
	end,
})

--ninja
animLoad('ninja', {
	sprites = {
		{file='ninja/ninja1.tga', bbox={-.75,-1.5,.75,.5}};		--stand
		{file='ninja/ninjac0.tga', bbox={-.75,-1.5,.75,.5}};	--pain / grab
		{file='ninja/ninja4.tga', bbox={-.75,-1.5,.75,.5}};		--run1
		{file='ninja/ninja3.tga', bbox={-.75,-1.5,.75,.5}};		--run2
		{file='ninja/ninja2.tga', bbox={-.75,-1.5,.75,.5}};		--run3 / jump / die
		{file='ninja/ninjaa0.tga', bbox={-.75,-1.5,.75,.5}};	--attack1
		{file='ninja/ninjaa1.tga', bbox={-.75,-1.5,.75,.5}};	--attack2
		{file='ninja/ninjar.tga', bbox={-.75,-1.5,.75,.5}};		--roll
	};
	seqs = {
		idle = {{joint=1; time=0; sprite=1}};
		pain = {{joint=1; time=0; sprite=2}};
		run = {
			{joint=1; time=0; sprite=3};
			{joint=1; time=.1; sprite=4};
			{joint=1; time=.2; sprite=5};
			{joint=1; time=.3; sprite=3};
		};
		jump = {{joint=1; time=0; sprite=5}};
		attack = {
			{joint=1; time=0; sprite=6};
			{joint=1; time=.2; sprite=7};
			{joint=1; time=.4; sprite=6};
		};
		roll = {{joint=1; time=0; sprite=8}};
		die = {{joint=1; time=0; sprite=5}};
	};
})

-- new hero
animLoad('jo', {
	sprites = {
		{file='jo/stand.tga', bbox={-1,-1.5,1,.5}};
		{file='jo/walk.tga', bbox={-1,-1.5,1,.5}};
	};
	seqs = {
		idle = {{joint=1; time=0; sprite=1}};
		pain = {{joint=1; time=0; sprite=2}};
		run = {
			{joint=1; time=0; sprite=2};
			{joint=1; time=.1; sprite=1};
			{joint=1; time=.2; sprite=2};
		};
		jump = {{joint=1; time=0; sprite=2}};
		fall = {{joint=1; time=0; sprite=1}};
		grab = {{joint=1; time=0; sprite=2}};
		attack = {{joint=1; time=0; sprite=2}};
		roll = {{joint=1; time=0; sprite=1}};
		die = {{joint=1; time=0; sprite=2}};
	};
})

-- this one is found in the physics module:
local baseFrictionScalar = 70

-- advection with post-fbo makes invisibility without the buffer swap slowdown
shader = renderShader{name='glass'; fragmentCode = advectionFragmentCode };

entclass('Player', TakesDamage, {

	bbox = box2:new(-.4, -1.4, .4, .4);
	animGroup = 'ninja';

	shader = renderShader{fragmentCode = invisibilityFragmentCode };

	ALLOW_PICKUP = false,
	ALLOW_BULLET_TIME = true,

	aimAngleRange = 85,
	aimAngleRate = 2,
	wallJumpThreshold = 0.5,
	wallJump = vec2:new(10),
	swimForce = 250,
	jumpVel = 21.5,
	runForce = 30,
	airForce = 15,
	runMaxVel = 10,
	painStun = 0.5,			-- how long you're stunned after pain
	painDelay = 1.5,		-- 
	flashlight = 0,

	jetpackHorzForce = vec2:new(40,40),
	jetpackVertForce = 60,
	jetpackIllum = vec3:new(1),
	jetpackIllumSize = vec2:new(15,4),
	jetpackRandCoeff = 0.7,
	
	illum = vec3:new(1),
	illumWidth = 12,
	illumHeight = 1.5,

	checkAttack = function(self)
		local doHook = false
		local doAttack = false

		if self.attackFinished < getTime() then
			if self.input.key.prev.isPressed then
				for i = 1,#weapons do
					self.weapon = self.weapon + #weapons - 1
					self.weapon = ((self.weapon-1) % #weapons) + 1
					if bitand(self.weaponFlags, bitshiftleft(1, self.weapon)) ~= 0 then break end
				end
			end
			if self.input.key.next.isPressed then
				for i = 1,#weapons do
					self.weapon = self.weapon + 1
					self.weapon = ((self.weapon-1) % #weapons) + 1
					if bitand(self.weaponFlags, bitshiftleft(1, self.weapon)) ~= 0 then break end
				end
			end
			if self.input.key.attack.isDown then
				doAttack = true
			end
		end
		
if Player.ALLOW_BULLET_TIME then

	local timescaleConvCoeff = 0.1	--time convergence constant
	local timescaleDest = 1
	if self.input.key.alt.isDown then timescaleDest = .1 end

	--dx/dt = c(x' - x)

	--Euler integration of constant difference equation (exponential convergence) over linear map
	--x(t+1) = x(t) + c * (x' - x(t))
	gameTimeScalar((1 - timescaleConvCoeff) * gameTimeScalar() + timescaleConvCoeff * timescaleDest);

end

		if doHook then
			weaponGrapplingHook.attackFunc(self, self.aimDir)
		end
		if doAttack then
--[[	attack anim doesn't look good enough to rob us of our lower body animation
			if	not self.input.key.left.isPressed and
				not self.input.key.right.isPressed and
				not self.input.key.up.isPressed
			then
				self:setAnimSeq('attack')
			end
--]]
			if self.weapon >= 1 and self.weapon <= #weapons then
				local w = weapons[self.weapon]
				w.attackFunc(self, self.aimDir)
				self.attackFinished = getTime() + w.attackDelay
			end
			if bitand(self.itemFlags, bitshiftleft(1, ITEM.FLIPOUT)) ~= 0 then
				self.attackFinished = -1
			end
		end
	end,
	
	updateTagsFromPos = function(self, pos)	-- void (entity, vec2f)
		local userTag = mapGetTag( math.floor(pos[1]), math.floor(pos[2]) )
		self:updateTags(userTag)
	end,
	
	updateTags = function(self, userTag)	-- void (entity, int)
		spawnObjSystemUpdateTags(userTag, self.tag)
		self.tag = userTag
	end,
	
	update = function(self)
		self:updateTagsFromPos( self:getPos() )
		if bitand(self.itemFlags, bitshiftleft(1, ITEM.JETPACK)) ~= 0 and getTime() > self.jetpackEndTime then
			self.itemFlags = bitand(self.itemFlags, bitnot(bitshiftleft(1, ITEM.JETPACK)))
		end
		if bitand(self.itemFlags, bitshiftleft(1, ITEM.FLIPOUT)) ~= 0 and getTime() > self.flipoutEndTime then
			--self.itemFlags = bitand(self.itemFlags, bitnot(bitshiftleft(1, ITEM.FLIPOUT)))
			self.itemFlags = self.itemFlags - bitshiftleft(1, ITEM.FLIPOUT)
		end
		if getTime() > self.screenFlashEndTime then
			self.screenFlash:set(0)
		end
	end,
	
	updateViewpos = function(self)
		local bounds = viewGetBox()
		local halfScreen = bounds:size()
		if halfScreen[1] >= -.00001 and halfScreen[1] <= .00001 and halfScreen[2] >= -.0001 and halfScreen[2] <= .0001 then 
			self.viewpos:set(self:getPos())
		else
			local mouseFrac = self.input.mouse.pos - .5
			local destPos = vec2:new(self:getPos()) + mouseFrac * halfScreen * 0.8
			self.viewpos:addTo((destPos - self.viewpos) * 0.0625)
		end
	end;
	
	think = function(self)
		Player.super.think(self)
		if not self:isActive() then return end	-- needed for any whose parent think() might kill themselves (Collidable)
		
		local mouseWorldPos = vec2:new(self.input.mouse.pos):map(viewGetBox())
		
		self:update()
		self:updateViewpos()
		
		if self.dead then return end
		
		if self.input.mouse.active then
			self.aimDir = (mouseWorldPos - self:getPos()):unit()
		else
			self.aimAngle = clamp(self.aimAngle, -self.aimAngleRange, self.aimAngleRange)
			self.aimDir = vec2.expd(self.aimAngle)
			if self:getLeft() then self.aimDir[1] = -self.aimDir[1] end
		end
		
if Player.ALLOW_PICKUP then
		if self.pickup then
			local center = vec2:new(self:getPosX(), self:getPosY() - 1)
			self.pickup:setVel(vec2:new(self:getVel()) + self.aimDir * 7.5)
			self.pickup:setPos(center + self.aimDir * 0.5)
		end
end
		if getTime() - self.lastPainTime < self.painStun then return end

if Player.ALLOW_PICKUP then
		if self.pickup then
			local o = self.pickup
			-- assert(o:isa(Collidable))
			if self.input.key.alt.isUp then
				o:setNoPhysics(false)
				if o:getSpawnObj() then
-- TODO - implement o:getSpawnObj()._ent as get/set spawnObjEnt
					if o:getSpawnObj()._ent then
						o:delete()
					else
						o:getSpawnObj()._ent = o
					end
				end
				self.pickup = nil
			end
		end
end
		
		if self.input.key.left.isPressed then self:setLeft(true) end
		if self.input.key.right.isPressed then self:setLeft(false) end
		
		if not self.input.mouse.active then
			if self.input.key.up.isDown then self.aimAngle = self.aimAngle - self.aimAngleRate end
			if self.input.key.down.isDown then self.aimAngle = self.aimAngle + self.aimAngleRate end
		else
			self:setLeft(mouseWorldPos[1] < self:getPosX())
		end
		
		if self:getContactNormalY() > -0.5 then
			if self:getLeft() and self:getContactNormalX() > 0.5 and
				getTime() - self:getContactTime() < self.wallJumpThreshold and
				getTime() - self.lastWallJumpTime > self.wallJumpThreshold
			then
				self:setAnimSeq('grab')
			elseif not self:getLeft() and self:getContactNormalX() < -0.5 and
				getTime() - self:getContactTime() < self.wallJumpThreshold and
				getTime() - self.lastWallJumpTime > self.wallJumpThreshold
			then
				self:setAnimSeq('grab')
			else
				self:setAnimSeq('jump')
			end
		else
			if self.attackFinished > getTime() then
--[[ i'm opting out of my attack anim.  let the weapons do the talking
				if self.input.key.left.isUp and
					self.input.key.right.isUp and
					self.input.key.up.isUp
				then
					self:setAnimSeq('attack')
				end
--]]
			else
				if self.input.key.left.isDown or
					self.input.key.right.isDown
				then
					self:setAnimSeq('run')
				else
					self:setAnimSeq('idle')
				end
			end
		end
		
		-- swimming
		
		if mapGetType(self:getPosX(), self:getPosY(), self:getLayer()) == MAPTYPE.Liquid then
			self:setAnimSeq('jump')	--swim
			local force = math.sin(getTime() * 2 * math.pi) * 0.5 + 0.5
			force = force * self.swimForce * getDeltaTime()
		
			if self.input.key.up.isDown then
				self:setVelY(self:getVelY() - force)
				
				if mapGetType(self:getPosX(), (self:getPosY() - 0.1), self:getLayer()) ~= MAPTYPE.Liquid then
					self:setVelY(-0.5 * self.jumpVel)
				end
			end
			if self.input.key.down.isDown then self:setVelY(self:getVelY() + force) end
			if self.input.key.left.isDown then self:setVelX(self:getVelX() - force) end
			if self.input.key.right.isDown then self:setVelX(self:getVelX() + force) end
			
		-- running
			
		elseif self:getContactNormalY() < -0.7 then
			if self.input.key.up.isDown then
				local n = (vec2:new(self:getContactNormal()) + vec2:new(0,-1)):unit()
				self:setVelX(self:getVelX() + n[1] * self.jumpVel)
				self:setVelY(self:getVelY() + n[2] * self.jumpVel)
			end
			if self.input.key.left.isDown then
				local grfr = mapGetFriction(self:getContactTileX(), self:getContactTileY(), self:getLayer(), SIDE.UP)
				local coeff = 2 - 2 / (grfr * grfr + 1)
				if self:getVelX() > -self.runMaxVel then
					self:setVelX(self:getVelX() - (self.runForce * coeff + baseFrictionScalar * grfr * self:getFriction()) * getDeltaTime())
					if self:getVelX() < -self.runMaxVel then
						self:setVelX(-self.runMaxVel)
					end
				end
			end
			if self.input.key.right.isDown then
				local grfr = mapGetFriction(self:getContactTileX(), self:getContactTileY(), self:getLayer(), SIDE.UP)
				local coeff = 2 - 2 / (grfr * grfr + 1)
				if self:getVelX() < self.runMaxVel then
					self:setVelX(self:getVelX() + (self.runForce * coeff + baseFrictionScalar * grfr * self:getFriction()) * getDeltaTime())
					if self:getVelX() > self.runMaxVel then
						self:setVelX(self.runMaxVel)
					end
				end
			end
			
		-- flight / wall movement
		
		else

			-- jetpack movement
		
			if bitand(self.itemFlags, bitshiftleft(1, ITEM.JETPACK)) ~= 0 then
				if self.input.key.up.isDown then
					if self.input.key.right.isDown then
						self:setVelX(self:getVelX() + self.jetpackHorzForce[1] * getDeltaTime())
						self:setVelY(self:getVelY() - self.jetpackHorzForce[2] * getDeltaTime())
					elseif self.input.key.left.isDown then
						self:setVelX(self:getVelX() - self.jetpackHorzForce[1] * getDeltaTime())
						self:setVelY(self:getVelY() - self.jetpackHorzForce[2] * getDeltaTime())
					else
						self:setVelY(self:getVelY() - self.jetpackVertForce * getDeltaTime())
					end
					
					local JETPACK_REFRESH = 0.1
					if math.floor(getTime() / JETPACK_REFRESH) ~= math.floor(getLastTime() / JETPACK_REFRESH) then
						JetpackFlame:new(self)
					end
					
					local lightPos = vec2:new()
					lightPos[2] = self:getPosY() - 1
					if self:getLeft() then lightPos[1] = self:getPosX() + 0.5
					else lightPos[1] = self:getPosX() - 0.5
					end
					local lightRot = vec2:new()
					if self:getLeft() then lightRot[1] = 1
					else lightRot[1] = -1
					end
					local lightScale = math.random() * self.jetpackRandCoeff + (1 - self.jetpackRandCoeff)
					lightPos = lightPos + lightRot * (self.jetpackIllumSize[1] * lightScale - 2)
					
					-- addLight params: pos, color, tex, range, [spot, rot]
					addLight(lightPos, self.jetpackIllum, spotLightTex,
						box2:new(
							-self.jetpackIllumSize[1] * lightScale,
							-self.jetpackIllumSize[2] * lightScale,
							self.jetpackIllumSize[1] * lightScale,
							self.jetpackIllumSize[2]
						),
						vec2:new(self:getPos()) - vec2:new(0,1),
						lightRot
					)
					
				else

					-- jetpack flying movement
				
					if self.input.key.left.isDown then
						if self:getVelX() > -self.runMaxVel then
							self:setVelX(self:getVelX() - self.airForce * getDeltaTime())
						end
					end
					if self.input.key.right.isDown then
						if self:getVelX() < self.runMaxVel then
							self:setVelX(self:getVelX() + self.airForce * getDeltaTime())
						end
					end
				end
				
			-- wall movement
			
			else
				if self:getContactNormalX() < -0.7 or self:getContactNormalX() > 0.7 then
					self.lastWallJumpTime = getTime() - self.wallJumpThreshold - 1
				end
				
				if self.input.key.left.isDown then
					if self:getContactNormalX() < -0.5 and
						getTime() - self:getContactTime() < self.wallJumpThreshold and
						getTime() - self.lastWallJumpTime > self.wallJumpThreshold 
					then
						self.lastWallJumpTime = getTime()
						self:setVelX(self:getVelX() - self.wallJump[1])
						self:setVelY(self:getVelY() - self.wallJump[2])
					else
						if self:getVelX() > -self.runMaxVel then
							self:setVelX(self:getVelX() - self.airForce * getDeltaTime())
						end
					end
				end

				if self.input.key.right.isDown then
					if self:getContactNormalX() > 0.5 and
						getTime() - self:getContactTime() < self.wallJumpThreshold and
						getTime() - self.lastWallJumpTime > self.wallJumpThreshold 
					then
						self.lastWallJumpTime = getTime()
						self:setVelX(self:getVelX() + self.wallJump[1])
						self:setVelY(self:getVelY() - self.wallJump[2])
					else
						if self:getVelX() < self.runMaxVel then
							self:setVelX(self:getVelX() + self.airForce * getDeltaTime())
						end
					end
				end
			end
		end
		
		self:checkAttack()
				
	end,
	
	touch = function(self,other)
		Player.super.touch(self,other)
		if self.input.key.alt.isUp then return end
		if self.pickup then return end
		
		if not other then return end
		
		if other:getNoPhysics() then return end
		other:setNoPhysics(true)

		if other:getSpawnObj() then
			other:getSpawnObj()._ent = nil
		end

		self.pickup = other
	end,
	
	respawn = function(self)
		local so = GameAPI:getStartSpawnObj()
		if not so then return end
		
		local sv = readVec2(so.pos)
		
		self:setAngle(0)
		self.dead = false
		self:setColor{1,1,1,1}
		self:init(sv)
	end,
	
	takeDamage = function(self, other, damage)
		if getTime() - self.lastPainTime < self.painDelay then return end
		Player.super.takeDamage(self, other, damage)
	end,
	
	pain = function(self, other)
		if getTime() - self.lastPainTime < self.painDelay then return end
		
		self.screenFlash:set(1,0,0,.25)
		self.screenFlashEndTime = getTime() + 0.05
		self.lastPainTime = getTime()
		
		self:setVelY(self:getVelY() - 10)
		if other and other:getPosX() < self:getPosX() then
			self:setVelX(self:getVelX() + 10)
		else
			self:setVelX(self:getVelX() - 10)
		end
		
		self:setAnimSeq('pain')
	end,
	
	death = function(self, other)
		if self.dead then return end
		
		self:setAngle(180)
		self:setColor{1,0,0,1}
		self.dead = true
		
		self:setAnimSeq('die')
		-- respawn after the animation sequence
		setTimeout(function() self:respawn() end, 5.0)
	end,

	init = function(self,pos)
		self:setPos(pos)
--		self:setBBox{min={-.4,-1.4}, max={.4,.4}}
		self:setFriction(1)
		self:setColor{1,1,1,1}

		self.viewpos = vec2:new(pos)
		self.health = 100
		self.maxHealth = self.health
		
		self.weapon = weaponSword.index		-- WEAPON.SWORD
		self.weaponFlags = bitshiftleft(1, self.weapon)
		
		self:setAnimGroup('ninja')
		self:setAnimSeq('idle')
		
		self.screenFlash = vec4:new()
		self.screenFlashEndTime = -1
		self.aimAngle = 0
		self.lastPainTime = 0
		self.attackFinished = 0
		self.lastWallJumpTime = 0
		self.itemFlags = 0

		self.jetpackEndTime = -1
		self.flipoutEndTime = -1
		
		-- for engine binding
		-- soon to be abstracted for all entities ...
		self.input = ClientInput:new()
		
		self:updateTags(-1)
	end,
	
	ctor = function(self, table)
		Player.super.ctor(self, table)
		self:init(self:getPos())
	end,
})
function Player.clientUpdate(classObj)
	Player.super.clientUpdate(classObj)
	drawHUD(net)

	if net.flashlight ~= 0 then
	
		local lightTex = circularIntenseLightTex
		if net.flashlight == 1 then lightTex = spotLightTex end

		local lightSpot = vec2:new(net.pos[1], net.pos[2] - 1)
		local rad = classObj.illumWidth
		if net.flashlight ~= 1 then rad = (classObj.illumWidth + classObj.illumHeight) * 0.5 end
		local lightRange = box2:new(-rad, rad)
		local lightRot = net.aimDir:copy()
		local lightPos = lightSpot:copy()
		if net.flashlight == 1 then
			self:setPos(vec2:new(net.pos) + net.aimDir * lightRange.max[1])
		end
		lightSpot = lightSpot + lightRot
		
		addLight(lightPos, classObj.illum, lightTex, lightRange, lightSpot, lightRot)
	end
end

-- append client render info
Player:createNetTable{
	aimDir = 'vec2f';
	viewpos = 'vec2f';
	health = 'float';
	maxHealth = 'float';
	screenFlash = 'color';
	weapon = 'int';
	weaponFlags = 'int';
	flipoutEndTime = 'float';
	jetpackEndTime = 'float';
	flashlight = 'int';
}

createClientStartClass(Player)
GameAPI.clientSpawnClass = PlayerStart

-- hack for keeping everything pretty
-- need to think of a better standard for this...
viewTileSize(32)

-- lets see how well overriding works
for _,cl in pairs(_entClasses) do
	cl.zScale = 0
	cl.zBias = -1.5
end
cvarExec('worldSpriteZScale 0')
cvarExec('worldTexZScale 0')
