require 'base.scripts.ents'
require 'ddiad.scripts.tile'
require 'ddiad.scripts.place'
require 'ddiad.scripts.vars'

class('Board')
function Board:ctor(args)
	self.size = args.size
	self.owner = args.owner
	self.isToric = args.toric or true
	self.places = table.new()		--2D array
	self.allPlaces = table.new()	--1D array
	for i=1,self.size do
		self.places[i] = table.new()
		for j=1,self.size do
			local p = Place:new()
			p.board = self
			p.isABoardPlace = true	-- false for hand places.  i could subclass i know...
			p.boardPos = vec2:new(i,j)
			p:setPos((p.boardPos - vec2:new(self.size/2)) * tileSize + self.owner:getPos())
			self.places[i][j] = p
			self.allPlaces:insert(p)
		end
	end
	self:fillWithColors()
end
function Board:fillWithColors()
	-- and fill it with random colors - make sure each color has 4 tiles
	local copy = table.new(self.allPlaces)
	local curcolor = 0
	while #copy > 0 do
		curcolor = curcolor % colorsUsed + 1
		local i = math.random(#copy)
		Tile:new(colors[curcolor]):setOn(copy[i])
		copy:remove(i)
	end
	do return end

	for i,c in ipairs(colors) do
		for i=1,#self.allPlaces / colorsUsed do
			local p
			repeat
				local allFilled = true
				for i,p in pairs(self.allPlaces) do
					if not p.tile then
						allFilled = false
						break
					end
				end
				assert(not allFilled, 'looks like we filled all our tiles before we should have')
				if allFilled then break end
				p = self.allPlaces[math.floor(math.random(#self.allPlaces))]
			until not p.tile
			Tile:new(c):setOn(p)
		end
	end
end

-- TODO - win-condition class!
function Board:hasWon()
	local checkColor
	for i,p in pairs(self.allPlaces) do
		if not p.tile then
			return false
		end
		if not p.tile.colorObj then
			return false
		end
		if not checkColor then
			checkColor = p.tile.colorObj
		else
			if p.tile.colorObj ~= checkColor then
				return false
			end
		end
	end
	return true
end
