---------------- party menu ----------------

-- todo - set onCancel to this
function menuCloseOnCancel(self, input)
	self:delete()
	return true
end

function menuCloseParent(self, input)
	self.parent:delete()
	return true
end

	-- serverside
	
-- if we put battle in its own location then maybe we should for this too

function TopviewClientView:openPartyMenu()
	fork(function()
		local args = {class='PartyMenu'}
		for charPartyIndex,charIndex in ipairs(self.clientInfo.chars) do
			args['char'..charPartyIndex] = charIndex
		end
		self.partyMenuOpen = true
		self:createPromptingMenu(args)
		self.partyMenuOpen = nil
	end)
end

function TopviewClientView:closePartyMenu()
	self:serverSendMessage('clientPartyMenuClose')
end

function TopviewClientView:serverPartyMenuGetItems()
	-- first send char equipment across the line
	--  'char' + index + field, item index, item name
	for charPartyIndex,charIndex in ipairs(self.clientInfo.chars) do
		local char = chars[charIndex]
		if char then
			for _,field in ipairs(equipFields) do
				local equippedItemIndex = char['equip'..field]
				if equippedItemIndex then
					local equippedItem = items[equippedItemIndex] 
					if equippedItem then
						self:serverSendMessage('clientPartyMenuGetItem', 'char'..charPartyIndex..field, equippedItemIndex, equippedItem.name)
					end
				end
			end
		end
	end
	-- next send inventory
	--  item index, item name, amount
	for _,itempair in ipairs(self.clientInfo.items) do
		local item = items[itempair.item]
		assert(item, "failed to find item for index "..tostring(itempair.item))
		self:serverSendMessage('clientPartyMenuGetItem', itempair.item, item.name, itempair.amount)
	end
	self:serverSendMessage('clientPartyMenuGetItem')
end

	-- clientside

class('ButtonMenu',Menu)
ButtonMenu.backgroundColor = {1,1,1,1}

class('ItemHolderMenu',ButtonMenu)
ItemHolderMenu.size={30,1.5}
ItemHolderMenu.isPrompt=true
function ItemHolderMenu:ctor(args)
	ItemHolderMenu.super.ctor(self, args)

	self.itemMenu = ButtonMenu:new{
		parent=self;
		pos={0,0};
		size={20, 1.5};
		text=args.itemName or '';
	}

	local amountText
	if self.itemAmount then
		assert(self.itemAmount ~= 0)
		if self.itemAmount > 0 then
			amountText = tostring(self.itemAmount)
		else
			amountText = ''
		end
	end
	self.amountMenu = ButtonMenu:new{
		parent=self;
		pos={20, 0};
		size={10, 1.5};
		text=amountText;
	}
end

class('ItemsMenu',Menu)
ItemsMenu.pos = {1,1}
ItemsMenu.size = Menu.sysSize - {2,2}
ItemsMenu.backgroundColor={1,1,1,1}
ItemsMenu.onCancel = menuCloseOnCancel
ItemsMenu.hasPrompts = true	-- this won't do anything unless there is at least one prompting child

function ItemsMenu:ctor(args)
	ItemsMenu.super.ctor(self, args)
	
	Menu:new{
		parent=self;
		pos={1, 1.5};
		size={10, 1.5};
		text='Arrange';
		isPrompt=true;
		onCancel=menuCloseParent;
	}
	
	for i,p in ipairs(self.client.partyMenuItems) do
		ItemHolderMenu:new{
			parent=self;
			pos={1, 1.5*(i+1)};
			size={20 ,1.5};
			itemIndex=p.index;
			itemName=p.name;
			itemAmount=p.amount;
			isPrompt=true;
			onCancel=menuCloseParent;
		}
	end
end

class('EquipMenu',Menu)
EquipMenu.pos = {1,1}
EquipMenu.size = Menu.sysSize - {4,4}
EquipMenu.backgroundColor={1,1,1,1}
EquipMenu.onCancel = menuCloseOnCancel

function EquipMenu:ctor(args)
	EquipMenu.super.ctor(self, args)
	self.client.equipMenu = self
	self:refresh()
end

function EquipMenu:onDelete()
	self.client.equipMenu = nil
end

--[[
this is a hack of a name.  onOk is the mouse click as well as the cursor punch when highlighted
this'll never be highlighted because it is a window, not a control, yet I have both under the 'Menu' class
--]]
function EquipMenu.childOnOk(selectedChild, input)
	local self = selectedChild.parent
	if selectedChild and selectedChild:isa(ItemHolderMenu) then
		if not self.client.toggledItemHolder then
			self.client.toggledItemHolder = selectedChild
			selectedChild.itemMenu.fontColor = {1, .5, 0, 1}
		else
			-- do a swap between self.client.toggledItemHolder
			-- and do a swap on the item structure that generated them
			local from = self.client.toggledItemHolder
			local to = selectedChild
			self.client.toggledItemHolder.itemMenu.fontColor = nil	-- clear the obj-based to fall back on class-based
			self.client.toggledItemHolder = nil
			
			--[[ now we swap 'from' and 'to'...
			is from or to is in the equip menu, and the other not? fail quietly if no
			is the to-be-equipped item compatible with this char? fail loud if no
			else do the swap
			--]]
			local fromLoc = from.equipField
			local toLoc = to.equipField

			-- don't forget that this is operating clientside
			-- that means whatever we do here we'll have to reflect on serverside
			
			if (fromLoc and not toLoc) or (toLoc and not fromLoc) then
				-- to will have the equip location, from will not (i.e. will be an item)
				if not toLoc then
					toLoc = fromLoc
					fromLoc = nil
					local tmp = from
					from = to
					to = tmp
				end
				-- now toLoc exists and fromLoc doesn't
				-- from is in the equip area and to isn't
				if from.itemIndex then
					entClientSendMessage(self.client.entID, 'serverEquipItem', self.charPartyIndex, toLoc, from.itemIndex)
				end
			end
		end
	end
end


function EquipMenu:refresh()
	if self.child then
		while #self.child > 0 do
			self.child[1]:delete()
		end
		self.child = nil
	end

	-- now for the equip pieces...
	-- head, body, lhand, rhand, relic1, relic2, esper
	
	self.equipHolders = table.new()
	for i,field in ipairs(equipFields) do
		ButtonMenu:new{
			parent=self;
			pos={1,1.5*i};
			size={10,1.5};
			text=field;
		};

		local charInfo = self.client.charEquipMenuItems[self.charPartyIndex]
		local itemIndex, itemName
		if charInfo and charInfo[field] then
			itemIndex = charInfo[field].index
			itemName = charInfo[field].name
		end

		self.equipHolders[field] = ItemHolderMenu:new{
			parent=self;
			pos={11, 1.5*i};
			itemIndex=itemIndex;
			itemName=itemName;
			amount=1;
			equipField=field;
			onOk=self.childOnOk;
			onCancel=menuCloseParent;
		};
		
		ButtonMenu:new{
			parent=self;
			pos={41, 1.5*i};
			size={1, 1.5};
			text='X';
			itemIndex=itemIndex;
			equipField=field;
			isPrompt=true;
			onOk=function(self)
				entClientSendMessage(self.client.entID, 'serverEquipItem', self.parent.charPartyIndex, self.equipField)
			end;
			onCancel=menuCloseParent;
		}
	end
	
	-- TODO - generic scrollable item list menu
	for i,p in ipairs(self.client.partyMenuItems) do
		ItemHolderMenu:new{
			parent=self;
			pos={1, 1.5*(i+#equipFields+1)};
			itemName=p.name;
			itemIndex=p.index;
			itemAmount=p.amount;
			onOk=self.childOnOk;
			onCancel=menuCloseParent;
		}
	end
end

function TopviewClientView:serverEquipItem(charPartyIndex, toLoc, fromItemIndex)
	fromItemIndex = tonumber(fromItemIndex)
	local fromItem
	if fromItemIndex then fromItem = items[fromItemIndex] end
	
	charPartyIndex = tonumber(charPartyIndex)
	local char = chars[self.clientInfo.chars[charPartyIndex]]
	if not char then return end

	-- test compatability
	if fromItem then
		if fromItem.type == 'weapon' then
			if toLoc ~= 'LHand' and toLoc ~= 'RHand' then return end
			-- TODO : test how many we're holding (one between L and R hand)
			-- ... or if we're holding a genji glove =P
		elseif fromItem.type == 'shield' then
			if toLoc ~= 'LHand' and toLoc ~= 'RHand' then return end
			-- TODO : test how many we're holding (one between L and R hand)
		elseif fromItem.type == 'helm' then
			if toLoc ~= 'Head' then return end
		elseif fromItem.type == 'armor' then
			if toLoc ~= 'Body' then return end
		end
		
		if fromItem.equipflags and bitand(fromItem.equipflags, char.equipflags) ~= fromItem.equipflags then
			return
		end
	end

	-- put it back
	do
		local toItemIndex = char['equip'..toLoc]
		if toItemIndex then
			local itemLocalIndex = table.find(self.clientInfo.items, toItemIndex, function(v, index) return v.item == index end)
			if itemLocalIndex then
				self.clientInfo.items[itemLocalIndex].amount = self.clientInfo.items[itemLocalIndex].amount + 1
			else
				table.insert(self.clientInfo.items, {item=toItemIndex, amount=1})
			end
		end
		char['equip'..toLoc] = nil
	end
	
	-- then put the from item in the to
	if fromItem then
		char['equip'..toLoc] = fromItemIndex
		local itemLocalIndex = table.find(self.clientInfo.items, fromItemIndex, function(v, index) return v.item == index end)
		assert(itemLocalIndex)
		self.clientInfo.items[itemLocalIndex].amount = self.clientInfo.items[itemLocalIndex].amount - 1
		if self.clientInfo.items[itemLocalIndex].amount == 0 then
			table.remove(self.clientInfo.items, itemLocalIndex)
		end
	end
	
	self:serverSendMessage('clientPartyMenuRefreshEquip')
end

function ClientSide:refreshPartyMenuItems()
	self.partyMenuItems = table.new()
	self.charEquipMenuItems = table.new()
	self.partyMenuItemsDone = nil
	entClientSendMessage(self.entID, 'serverPartyMenuGetItems')
	while not self.partyMenuItemsDone do sleep(0) end
	self.partyMenuItemsDone = nil
end

function ClientSide:clientPartyMenuRefreshEquip()
	fork(function()
		-- first refresh the clientside item list we have stored ...
		-- if I do sql this sucker, should the client have access to database reads?
		-- nahh...
		self:refreshPartyMenuItems()
	
		-- next refresh the EquipMenu contents...
		self.equipMenu:refresh()
	end)
end


class('CharOptionsMenu',Menu)
CharOptionsMenu.size={10, 7}
CharOptionsMenu.isPrompt=true
function CharOptionsMenu:ctor(args)
	CharOptionsMenu.super.ctor(self,args)
	ButtonMenu:new{
		parent=self;
		pos={1, .5};
		size={5, 1.5}; 
		text='Equip';
		isPrompt=true;
		onOk=function(self)
			-- TODO - spawn only one?  attch and detach, show and hide, instead?
			-- force z order?
			local e = EquipMenu:new{
				parent=self.client.partyMenu;
				charIndex=self.parent.charIndex;
				charPartyIndex=self.parent.charPartyIndex;
				hasPrompts=true;
			}

			self.parent:delete()
		end;
		onCancel=menuCloseParent;
	}
	ButtonMenu:new{parent=self; pos={1, 2.5}; size={5, 1.5}; text='Skills';}
	ButtonMenu:new{parent=self; pos={1, 4.5}; size={5, 1.5}; text='Stats';}
end
function CharOptionsMenu:onDelete()
	if self == self.client.charOptionsMenu then
		self.client.charOptionsMenu = nil
	end
end

class('CharMenu',Menu)
CharMenu.size = {40, 7}
CharMenu.isPrompt=true
function CharMenu:ctor(args)
	CharMenu.super.ctor(self,args)
	
	ButtonMenu:new{
		parent=self;
		pos={0,0};
		size={self.size[2], self.size[2]};
		backgroundColor = {1,1,1,1};
		-- NOTICE - here's a ref to the serverside sprite index ...
		-- what to do? request / map / cache on the clientside?
		tex = chars[self.charIndex].menuSprite;
	}
	
	-- TODO - serverside request
	local char = chars[self.charIndex]

	local y = .5
	Menu:new{
		parent=self;
		pos={self.size[2]+1, y};
		size={10,1.5};
		text=char.name;
	}
	y=y+1.5

	Menu:new{
		parent=self;
		pos={self.size[2]+1, y};
		size={10,1.5};
		text='Level '..char.level;
	}
	y=y+1.5
	
	Menu:new{
		parent=self;
		pos={self.size[2]+1, y};
		size={10,1.5};
		text='HP: '..char.hp..'/'..char.maxHP;
	}
	y=y+1.5

	Menu:new{
		parent=self;
		pos={self.size[2]+1, y};
		size={10,1.5};
		text='MP: '..char.mp..'/'..char.maxMP;
	}
	y=y+1.5
end

function CharMenu:onOk()
	-- pop up char sub menu stuff beside them here...
	--	Equip (combine with Relic)
	--	Skills
	--	Status
	if self.client.charOptionsMenu then self.client.charOptionsMenu:delete() end
	self.client.charOptionsMenu = CharOptionsMenu:new{
		parent=self.client.partyMenu;
		pos={self.pos[1] + self.size[2] + 1, self.parent.pos[2]};
		charIndex=self.charIndex;
		charPartyIndex=self.charPartyIndex;
		hasPrompts=true;
	}
end;

class('PartyMenu',Menu)
PartyMenu.pos = {1,1}
PartyMenu.size = Menu.sysSize - {2,2}
PartyMenu.onCancel = menuCloseOnCancel
PartyMenu.hasPrompts = true

function PartyMenu:onDelete()
	self.client.partyMenu = nil
end

function PartyMenu:ctor(args)
	PartyMenu.super.ctor(self, args)
	if self.client.partyMenu then
		self.client.partyMenu:delete()
	end
	self.client.partyMenu = self

	--don't call wait clientside...
	fork(function()
		self.client:refreshPartyMenuItems()

		for i=1,4 do
			-- chars being a global that keeps track of all unique characters ...
			local charIndex = args['char'..i]
			if charIndex then
				charIndex = tonumber(charIndex)
				CharMenu:new{
					parent=self;
					pos={1, (CharMenu.size[2]+1)*(i-1)+1};
					charPartyIndex=i;
					charIndex=charIndex;
					onCancel=menuCloseParent;
				}
			end
		end

		for i,v in ipairs{'Items', 'Config'} do
			ButtonMenu:new{
				parent=self;
				pos={self.size[1]-10, 2*i};
				size={9,1.5};
				text=v;
				onOk=function(self, input)
					local cl = _G[v..'Menu']	-- i.e. ItemMenu
					if cl then cl:new{client=self.client} end
				end;
				isPrompt=true;
				onCancel=menuCloseParent;
			}
		end
		
	end)
end

-- two overloads for the clientPartyMenuGetItem method
-- first is 'char' + index + field, item index, item name
-- next is item index, item name, amount
function ClientSide:clientPartyMenuGetItem(...)
	local args = {...}
	if #args == 0 then
		self.partyMenuItemsDone = true
	elseif args[1]:sub(1,4) == 'char' then
		local charPartyIndex, field = args[1]:match('^char(%d+)(.*)$')
		charPartyIndex = assert(tonumber(charPartyIndex))
		local index = assert(tonumber(args[2]))
		local name = assert(args[3])
		assert(field)

		assert(self.charEquipMenuItems)
		if not self.charEquipMenuItems[charPartyIndex] then
			self.charEquipMenuItems[charPartyIndex] = {}
		end
		self.charEquipMenuItems[charPartyIndex][field] = {
			index=index; 
			name=name;
		}
	else
		local index = assert(tonumber(args[1]))
		local name = assert(args[2])
		local amount = assert(tonumber(args[3]))
		assert(self.partyMenuItems):insert{
			index=index;
			name=name;
			amount=amount;
		}
	end
end

function ClientSide:clientPartyMenuClose()
	if self.partyMenu then 
		self.partyMenu:delete()
	end
end

