require 'wellstory.scripts.states'

animLoad('bat', 'bat/bat.anim')

entclass('Bat', StateEnt)

Bat.animGroup = 'bat'
Bat.bbox = box2:new(.45)
	
Bat.backToWaitingTime = 1.3
Bat.killdir = vec2:new()

Bat.STATE = {}

-- when the bat is doing nothing in particular
-- waiting for more BLOOOOD to DRINK
Bat.STATE.WAIT = {}
function Bat.STATE.WAIT:enter()
	self:setAngle(180)
end
function Bat.STATE.WAIT:think()
	--if the player is close enuf then flap towards them
	for ient in allents() do
		
		if ient and
			ient:isa(Player) and
			ient.health > 0 and
			vec2.distSq(ient:getPos(), self:getPos()) < self.wakeDist*self.wakeDist
		then
			-- wake up the bat, enter flapping state, and fly towards the player
			self.enemy = ient
			self:setState(self.STATE.SEEK)
			self.killdir = vec2:new(self.enemy:getPos()):subTo(self:getPos()):unit()
			break
		end
	end
end
function Bat.STATE.WAIT:leave()
	self:setAngle( 0)
end
		
		-- SEEKING after the player
Bat.STATE.SEEK = {}
function Bat.STATE.SEEK:enter()
	self:setAnimSeq('run')
end
function Bat.STATE.SEEK:think()
	if not self.enemy or self.enemy.health <= 0 then
		self.enemy = nil
		self:setState(self.STATE.WAIT)
	else
		self:setVel(self.killdir:copy())
		self:setVelY(self:getVelY() + 5 * math.pow(math.sin(5 * getTime()), 10))
		self:setVel(vec2:new(self:getVel()):unit() * 5)		--self.speed
	end
end
function Bat.STATE.SEEK:touch(other)
	self:setState(self.STATE.BACKTOWAITING)
end
function Bat.STATE.SEEK:leave()
	self:setAnimSeq('idle')
end
		
Bat.STATE.BACKTOWAITING = {}
function Bat.STATE.BACKTOWAITING:enter()
	self:setAnimSeq( 'run')
	self:setAngle(180)
	self:setVel{0,-5}

	self.enemy = nil
	self.killdir = nil
	self.stateStartTime = getTime()
end
function Bat.STATE.BACKTOWAITING:think()
	local stateTime = getTime() - self.stateStartTime
	if stateTime > self.backToWaitingTime then
		self:setState(self.STATE.ALMOSTWAITING)
		return
	end
end
function Bat.STATE.BACKTOWAITING:touch(other)
	if not other and self:getContactNormalY() > 0 then
		self:setState(self.STATE.ALMOSTWAITING)
	end
end
function Bat.STATE.BACKTOWAITING:leave()
	self:setAngle(0)
	self:setAnimSeq('idle')
end

Bat.STATE.ALMOSTWAITING = {}
-- fall down a little ...
function Bat.STATE.ALMOSTWAITING:enter()
	self:setAngle( 180)
	self:setVel{0,.1}
	self.stateStartTime = getTime()	--soon enuf this'll go in the __newindex code
end

-- then next chance you get (timer anyone) go to wait state
function Bat.STATE.ALMOSTWAITING:think()
	local stateTime = getTime() - self.stateStartTime
	if stateTime > self.almostWaitingTime then
		self:setVel{0,0}
		self:setState(self.STATE.WAIT)
	end
end
			
function Bat.STATE.ALMOSTWAITING:leave() 
	self:setAngle(0)
end

Bat.STATE.DEAD = {}
--nothing to see here folks ...
--removeme after a while

Bat.maxHealth = 1	--TakesDamage

Bat.wakeDist = 6
Bat.almostWaitingTime = 0.2

Bat.lastPainTime = 0	--defaults
Bat.painStun = 0.5

-- todo: do without?
function Bat:ctor(so)
	Bat.super.ctor(self, so)
	
	-- how do we make something touchable but not collidable?
	self:setTouchOnly(true)	--what good is touch only anyways?
	self:setUseGravity(false)
	self:setFriction(0)
	
	self:setState(self.STATE.WAIT)
end
	
function Bat:touch(other)
	Bat.super.touch(self,other)
	
	if self.state and self.state.touch then self.state.touch(self,other) end
	
	-- hmm ... 
	if other and other:isa(Player) and other.health > 0 then
		other:takeDamage(self, 1)
	end
end
	
function Bat:think()
	Bat.super.think(self)
	if not self:isActive() then return end	-- needed for any whose parent think() might kill themselves (Collidable)

	if getTime() - self.lastPainTime < self.painStun then return end
	self:setUseGravity(false)	--incase itw as changed... as pain will do
	
	--moved to super.  should it be called at beginning or end?
	--if self.state and self.state.think then self.state.think(self) end
end

function Bat:pain(other)
	self.lastPainTime = getTime()

	self:setVelY(self:getVelY() - 10)
	if other and other:getPosX() < self:getPosX() then
		self:setVelX(self:getVelX() + 10)
	else
		self:setVelX(self:getVelX() - 10)
	end
	
	self:setUseGravity(true)
end
	
function Bat:death(other, damage)
	self:setState(self.STATE.DEAD)
	self:setTouchOnly(false)
	self:setIgnoreEnts(true)
	self:setUseGravity(true)
	Bat.super.death(self,other,damage)	--make him spin and go away
	-- why isnt this working in the superclass?
	setTimeout( function() self:delete() end, 1)
end

Spawnable(Bat)
