require 'base.scripts.lighttex'
require 'base.scripts.clientinput'
require 'base.scripts.client'
require 'base.scripts.defs'
require 'wellstory.scripts.text'	--floating text, and ShowsDamage
require 'wellstory.scripts.hud'
require 'base.scripts.gameapi'

animLoad('player', 'mk/mk.anim')

-- this one is found in the physics module:
local baseFrictionScalar = 70

entclass('Player', ShowsDamage,{

	aimAngleRange = 85,
	aimAngleRate = 2,

	wallJumpThreshold = 0.5,
	wallJump = vec2:new(10),
	ALLOW_WALLJUMP = false,

	swimForce = 250,
	jumpVel = 21.5,
	jumpFromWaterVel = 21.5;
	runForce = 30,
	airForce = 15,
	runMaxVel = 10,
	painStun = 0.5,			-- how long you're stunned after pain
	painDelay = 1.5,		-- 
	flashlight = 0,

	illum = vec3:new(1),
	illumWidth = 12,
	illumHeight = 1.5,

	checkAttack = function(self)
		local doAttack = false

		if self.attackFinished < getTime() then
			if self.input.key.prev.isPressed then
				for i = 1,#weapons do
					backpack.weapon = backpack.weapon + #weapons - 1
					backpack.weapon = ((backpack.weapon-1) % #weapons) + 1
					if bitand(backpack.weaponFlags, bitshiftleft(1, backpack.weapon)) ~= 0 then break end
				end
			end
			if self.input.key.next.isPressed then
				for i = 1,#weapons do
					backpack.weapon = backpack.weapon + 1
					backpack.weapon = ((backpack.weapon-1) % #weapons) + 1
					if bitand(backpack.weaponFlags, bitshiftleft(1, backpack.weapon)) ~= 0 then break end
				end
			end
			if self.input.key.attack.isDown then
				doAttack = true
			end
		end
		if doAttack then
			if	not self.input.key.left.isPressed and
				not self.input.key.right.isPressed and
				not self.input.key.up.isPressed
			then
				self:setAnimSeq('attack')
			end
			if backpack.weapon >= 1 and backpack.weapon <= #weapons then
				local w = weapons[backpack.weapon]
				w.attackFunc(self, self.aimDir)
				self.attackFinished = getTime() + w.attackDelay
			end
		end
	end;

	updateTags = function(self, userTag)	-- void (entity, int)
		spawnObjSystemUpdateTags(userTag, self.tag)
		self.tag = userTag
	end;

	updateTagsFromPos = function(self, pos)	-- void (entity, vec2f)
		local userTag = mapGetTag( math.floor(pos[1]), math.floor(pos[2]) )
		self:updateTags(userTag)
	end,
	
	update = function(self)
		self:updateTagsFromPos( vec2:new(self:getPos()) )
		if getTime() > self.screenFlashEndTime then
			self.screenFlash:set(0)
		end
	end,
	
	updateViewpos = function(self)
		-- if this is a client method then we need to expand on net tables for this to work
		if not self.disableInput then
			if self.input.mouse.active then
				local bounds = viewGetBox()
				local halfScreen = bounds:size()
				local mouseFrac = self.input.mouse.pos - .5
				self.destViewpos:set(mouseFrac):mulTo(halfScreen):mulTo(0.8):addTo(self:getPos())
			else 
				self.destViewpos:set(self:getPos())
			end
		end

		self.viewpos:addTo( (self.destViewpos - self.viewpos) * 0.0625 )
	end;
	
	think = function(self)
		Player.super.think(self)
		if not self:isActive() then return end	-- needed for any whose parent think() might kill themselves (Collidable)
		
		local mouseWorldPos = vec2:new(self.input.mouse.pos):map(viewGetBox())

		self:update()
		self:updateViewpos()
		
		if self.dead then return end
		
		if self.input.mouse.active then
			self.aimDir = (mouseWorldPos - self:getPos()):unit()
		else
			self.aimAngle = clamp(self.aimAngle, -self.aimAngleRange, self.aimAngleRange)
			self.aimDir = vec2.expd(self.aimAngle)
			if self:getLeft() then self.aimDir[1] = -self.aimDir[1] end
		end
		
		-- if we went from inside water to outside water then make the player hop up a block's worth
		local newMapType = mapGetType(self:getPosX(), self:getPosY(), self:getLayer())
		if newMapType ~= self.lastMapType then
			if self.lastMapType == MAPTYPE.Liquid and newMapType ~= MAPTYPE.Liquid then
				if self:getVelY() > -self.jumpFromWaterVel then
					self:setVelY(-self.jumpFromWaterVel)
				end
			end
		end
		self.lastMapType = newMapType
		
		if self.pickup then
			local center = vec2:new(0,-1) + self:getPos()
			self.pickup:setVel(self:getVel() + self.aimDir * 7.5)
			self.pickup:setPos(center + self.aimDir * 0.5)
		end

		if getTime() - self.lastPainTime < self.painStun then return end

		if self.pickup then
			local o = self.pickup
			-- assert(o:isa(Collidable))
			if self.input.key.attack.isUp then
				o:setNoPhysics(false)
				if o:getSpawnObj() then
					if o:getSpawnObj()._ent then
						o:delete()
					else
						o:getSpawnObj()._ent = o
					end
				end
				self.pickup = nil
			end
		end
		
		--recent hack
		-- try to cut the wire when the script says so
		if self.disableInput then
			self.input:setcmd(0)
			self.input.mouse.active = false
		end
		
		if self.input.key.left.isPressed then self:setLeft(true) end
		if self.input.key.right.isPressed then self:setLeft(false) end
		
		if not self.input.mouse.active then
			if self.input.key.up.isDown then self.aimAngle = self.aimAngle - self.aimAngleRate end
			if self.input.key.down.isDown then self.aimAngle = self.aimAngle + self.aimAngleRate end
			if self.input.key.left.isDown then self.aimAngle = self.aimAngle * 0.9 end
			if self.input.key.right.isDown then self.aimAngle = self.aimAngle * 0.9 end
		else
			self:setLeft(mouseWorldPos[1] < self:getPosX())
		end
		
		if self:getContactNormalY() > -0.5 and self.ALLOW_WALLJUMP then
			if self:getLeft() and self:getContactNormalX() > 0.5 and
				getTime() - self:getContactTime() < self.wallJumpThreshold and
				getTime() - self.lastWallJumpTime > self.wallJumpThreshold
			then
				self:setAnimSeq('grab')
			elseif not self:getLeft() and self:getContactNormalX() < -0.5 and
				getTime() - self:getContactTime() < self.wallJumpThreshold and
				getTime() - self.lastWallJumpTime > self.wallJumpThreshold
			then
				self:setAnimSeq('grab')
			else
				self:setAnimSeq('jump')
			end
		else
			if self.attackFinished > getTime() then
				if self.input.key.left.isUp and
					self.input.key.right.isUp and
					self.input.key.up.isUp
				then
					self:setAnimSeq('attack')
				end
			else
				if self.input.key.left.isDown or
					self.input.key.right.isDown
				then
					self:setAnimSeq('run')
				else
					self:setAnimSeq('idle')
				end
			end
		end
		
		-- swimming
		
		if mapGetType(self:getPosX(), self:getPosY(), self:getLayer()) == MAPTYPE.Liquid then
			self:setAnimSeq('jump')	--swim
			local force = math.sin(getTime() * 2 * math.pi) * 0.5 + 0.5
			force = force * self.swimForce * getDeltaTime()
		
			if self.input.key.up.isDown then
				self:setVelY(self:setVelY() - force)
				
				if mapGetType(self:getPosX(), (self:getPosY() - 0.1), self:getLayer()) ~= MAPTYPE.Liquid then
					self:setVelY(-0.5 * self.jumpVel)
				end
			end
			if self.input.key.down.isDown then self:setVelY(self:getVelY() + force) end
			if self.input.key.left.isDown then self:setVelX(self:getVelX() - force) end
			if self.input.key.right.isDown then self:setVelX(self:getVelX() + force) end
			
		-- running
			
		elseif self:getContactNormalY() < -0.7 then
			if self.input.key.up.isDown then
				local n = (self:getContactNormal() + vec2:new(0,-1)):unit()
				self:setVelX(self:getVelX() + n[1] * self.jumpVel)
				self:setVelY(self:getVelY() + n[2] * self.jumpVel)
			end
			if self.input.key.left.isDown then
				local grfr = mapGetFriction(self:getContactTileX(), self:getContactTileY(), self:getLayer(), SIDE.UP)
				local coeff = 2 - 2 / (grfr * grfr + 1)
				if self:getVelX() > -self.runMaxVel then
					self:setVelX(self:getVelX() - (self.runForce * coeff + baseFrictionScalar * grfr * self:getFriction()) * getDeltaTime())
					if self:getVelX() < -self.runMaxVel then
						self:setVelX(-self.runMaxVel)
					end
				end
			end
			if self.input.key.right.isDown then
				local grfr = mapGetFriction(self:getContactTileX(), self:getContactTileY(), self:getLayer(), SIDE.UP)
				local coeff = 2 - 2 / (grfr * grfr + 1)
				if self:getVelX() < self.runMaxVel then
					self:setVelX(self:getVelX() + (self.runForce * coeff + baseFrictionScalar * grfr * self:getFriction()) * getDeltaTime())
					if self:getVelX() > self.runMaxVel then
						self:setVelX(self.runMaxVel)
					end
				end
			end
			
		-- flight / wall movement
		
		else
			-- wall movement
			
			if self.ALLOW_WALLJUMP then
				if self:getContactNormalX() < -0.7 or self:getContactNormalX() > 0.7 then
					self.lastWallJumpTime = getTime() - self.wallJumpThreshold - 1
				end
			end
			
			-- flight / wall movement
				
			if self.input.key.left.isDown then
				if self.ALLOW_WALLJUMP and self:getContactNormalX() < -0.5 and
					getTime() - self:getContactTime() < self.wallJumpThreshold and
					getTime() - self.lastWallJumpTime > self.wallJumpThreshold 
				then
					self.lastWallJumpTime = getTime()
					self:setVelX(self:getVelX() - self.wallJump[1])
					self:setVelY(self:getVelY() - self.wallJump[2])
				else
					if self:getVelX() > -self.runMaxVel then
						self:setVelX(self:getVelX() - self.airForce * getDeltaTime())
					end
				end
			end

			if self.input.key.right.isDown then
				if self.ALLOW_WALLJUMP and self:getContactNormalX() > 0.5 and
					getTime() - self:getContactTime() < self.wallJumpThreshold and
					getTime() - self.lastWallJumpTime > self.wallJumpThreshold 
				then
					self.lastWallJumpTime = getTime()
					self:setVelX(self:getVelX() + self.wallJump[1])
					self:setVelY(self:getVelY() - self.wallJump[2])
				else
					if self:getVelX() < self.runMaxVel then
						self:setVelX(self:getVelX() + self.airForce * getDeltaTime())
					end
				end
			end
		end
		
		self:checkAttack()
	end,
	
	touch = function(self,other)
		Player.super.touch(self,other)
		if not other then return end
		if other:getNoPhysics() then return end
		if not other.canBePickedUp then return end	-- here's our new interface for pickupable's
		if self.input.key.attack.isUp then return end
		if self.pickup then return end
		other:setNoPhysics(true)

		if other:getSpawnObj() then
			other:getSpawnObj()._ent = nil
		end

		self.pickup = other
	end,
	
	respawn = function(self)
		local so = GameAPI:getStartSpawnObj()
		if not so then return end
		
		local sv = readVec2(so.pos)
		
		self:setAngle(0)
		self.dead = false
		self:setColor{1,1,1,1}
		self:init(sv)
	end,
	
	takeDamage = function(self, other, damage)
		if getTime() - self.lastPainTime < self.painDelay then return end
		-- when the player takes damage, have a minus sign with some numbers float up above his head
		-- should we do this with enemies too?
		--superclass!
		Player.super.takeDamage(self, other, damage)
	end,
	
	pain = function(self, other)
		if getTime() - self.lastPainTime < self.painDelay then return end
		
		self.screenFlash:set(1,0,0,.25)
		self.screenFlashEndTime = getTime() + 0.05
		self.lastPainTime = getTime()
		
		self:setVelY(self:getVelY() - 10)
		if other and other:getPosX() < self:getPosX() then
			self:setVelX(self:getVelX() + 10)
		else
			self:setVelX(self:getVelX() - 10)
		end
		
		self:runAnimSeq('pain')
	end,
	
	death = function(self, other)
		if self.dead then return end
		
		self:setAngle(180)
		self:setColor{1,0,0,1}
		self.dead = true
		
		self:setAnimSeq('die')
		-- respawn after the animation sequence
		setTimeout(function() self:respawn() end, 5.0)
	end,

	init = function(self,pos)
		self:setPos(pos)
		self.viewpos = vec2:new(pos)
		self.destViewpos = vec2:new(pos)
		self:setBBox{min={-.4,-1.4},max={.4,.4}}
		
		-- my friction model SUXXORS
		self:setFriction(1)

		self.health = 5
		self.maxHealth = self.health
		self:setColor{1,1,1,1}
		
		self:setAnimGroup('player')
		self:setAnimSeq('idle')
		
		self.screenFlash = vec4:new()
		self.screenFlashEndTime = -1
		self.aimAngle = 0
		self.lastPainTime = 0
		self.attackFinished = 0
		self.lastWallJumpTime = 0

		-- for engine binding
		-- soon to be abstracted for all entities ...
		self.input = ClientInput:new()
		
		self:updateTags(-1)
	end,
	
	ctor = function(self, table)
		Player.super.ctor(self, table)
		self:init(self:getPos())
	end,
})
function Player.clientUpdate(classObj)
	Player.super.clientUpdate(classObj)
	
	drawHUD(net)
	if net.flashlight then
	
		local lightTex = circularIntenseLightTex
		if net.flashlight == 1 then lightTex = spotLightTex end

		local lightSpot = vec2:new(net.pos[1], net.pos[2] - 1)
		local rad = classObj.illumWidth
		if net.flashlight ~= 1 then rad = (classObj.illumWidth + classObj.illumHeight) * 0.5 end
		local lightRange = box2:new(-rad, rad)
		if net.flashlight == 1 then
			lightRange.min[1] = lightRange.min[2] * 0.5
			lightRange.max[2] = lightRange.max[2] * 0.5
		end
		local lightRot = vec2:new(net.aimDir)
		local lightPos = lightSpot:copy()
		if net.flashlight == 1 then
			lightPos:addTo(lightRot * (lightRange.max[1] - 1))
		end
		lightSpot:addTo(lightRot)
		
		addLight(lightPos, classObj.illum, lightTex, lightRange, lightSpot, lightRot)
	end

	-- here's our update-the-clientside command,
	-- needs to be executed by whatever class is gonna handle clientside stuff
	ClientSide:getClient():update()
end


function Player:serverRecvMessage(cmd, ...)
	local f = self[cmd]
	if not f then
		print('Player:serverRecvMessage got an unknown command ',cmd)
	else
		f(self, ...)
	end
end

function Player.clientRecvMessage(cmd, ...)
	local cl = ClientSide:getClient()
	local f = cl[cmd]
	if not f then
		print('Player.clientRecvMessage got an unknown command ',cmd)
	else
		f(cl, ...)
	end
end


Player:createNetTable{
	viewpos = 'vec2f';
	health = 'float';
	maxHealth = 'float';
	screenFlash = 'color';
	flashlight = 'bool';
	aimDir = 'float';
}

createClientStartClass(Player, 'player', box2:new(-.4, -1.4, .4, .4))
GameAPI.clientSpawnClass = PlayerStart

makeEntityMenuCapable(Player)

-- hack for keeping everything pretty
-- need to think of a better standard for this...
viewTileSize(32)

cvarExec('worldSpriteZScale 0')
cvarExec('worldTexZScale 0')
