require 'wellstory.scripts.ents'
require 'base.scripts.util'
require 'base.scripts.vec'
require 'base.scripts.lighttex'

-- TODO - think this through better & cleaner

-- collect all moltons!
weapons = table.new()

-- returns a table with the correct constructions
entclass('WeaponClass')
function WeaponClass:ctor(name, groupName, spawnName, attackDelay, attackFunc, renderFunc)
	self.name = name
	self.groupName = groupName		-- string
	self.spawnName = spawnName		-- string
	self.attackDelay = attackDelay	-- number
	self.renderFunc = renderFunc	-- void (entity,vec2)
	self.attackFunc = attackFunc	-- same

	weapons:insert(self)	-- numeric mapping
	weapons[name] = self	-- name mapping
	
	self.index = #weapons	-- used for mapping to weaponFlags (in client render)
	self.weaponFlag = bitshiftleft(1, self.index)
end


entclass('EffectLine', Renderable)

EffectLine.duration = 0.1

	--start, rot, dist, thick, tex, color, tc, use_lighting
function EffectLine:ctor(start, rot, dist, thick, tex, color, tc, useLighting)
	EffectLine.super.ctor(self)
	self.start = start:copy()
	self.rot = rot:copy()
	self.dist = dist
	self.thick = thick
	self.tex = tex
	self:setColor(color)
	self.tc = tc:copy()
	self.useLighting = useLighting
	self.createTime = getTime()
end
	
function EffectLine:think()
	if getTime() > self.createTime + self.duration then
		self:delete()
		return
	end
end

function EffectLine.clientUpdate(classObj)
	drawLine(
		net.start,
		net.rot,
		net.dist,
		net.thick,
		net.tex,
		net.color,
		net.tc,
		net.useLighting)
end

EffectLine:createNetTable{
	start = 'vec2f';
	rot = 'vec2f';	-- replace with use of angle?
	dist = 'float';
	thick = 'float';
	tex = 'uint';
	tc = 'vec2f';
	useLighting = 'bool';
}

-- void (vec2, int)
function Explosion(pos, parCount)
	for j = 1,parCount do
		local shade = math.random()
		local scalar = math.random()
		Particle:new{
			pos = pos;
			vel = vec2.exp( math.random() * 2 * math.pi) * ((1 - scalar) * (1 - scalar));
			radius = scalar * 4;
			color = vec4:new(1,shade,shade*0.5,1);
			life = 2 * (1 - scalar) * (1 - scalar) + math.random() * 0.3 + 0.3;
		}
	end
end

-- void (vec2, float, entity, entity)
function RadiusDamage(pos, radius, dmg, source, ignore)
	for ient in allents() do
		if ient and
			ient:isa(TakesDamage) and
			ient ~= ignore and
			ient.health > 0
		then
			local del = vec2:new(ient:getPos()) - pos
			local mag = del:len()
			if mag > 0 and mag < radius then
				del = del * (20 * ient:getMassInv() / mag)
				ient:setVel(vec2:new(ient:getVel()) + del)
				ient:takeDamage(source, dmg)
			end
		end
	end
end



-------------------------- SWORD --------------------------

animLoad('sword', 'swingsword.png')

entclass('SwordLight', Renderable)
SwordLight.duration = 2.5
SwordLight.illum = vec3:new(5,5,1,1)
SwordLight.illumBox = box2:new(-10,10)

function SwordLight:ctor(sword)
	SwordLight.super.ctor(self)
	self:setPos(sword:getPos())
	self.createTime = getTime()
end
	
function SwordLight:think()
	local frac = (getTime() - self.createTime) / self.duration
	
	if frac < 0 or frac > 1 then
		self:delete()
		return
	end
end
	
function SwordLight.clientUpdate(classObj)
	SwordLight.super.clientUpdate(classObj)
	-- addLight params: pos, color, tex, range, [spot, rot]
	local frac = (getTime() - net.createTime) / classObj.duration
	local lum = 1 - frac
	lum = lum * lum * lum	-- (1-frac)^3
	addLight(net.pos, classObj.illum * lum, circularLightTex, classObj.illumBox)
end

SwordLight:createNetTable{
	createTime = 'float';
}


entclass('Sword', Renderable)
Sword.angleRange = 70
Sword.range = 3
Sword.force = vec2:new(20,10)	-- 30 20
Sword.damage = 1
	
function Sword:ctor(e, dir)
	Sword.super.ctor(self)
	self.createTime = getTime()
	self.duration = weaponSword.attackDelay		-- weapons[WEAPON.SWORD].attackDelay
	self.holder = e
	
	if not e:getLeft() then
		self:setAngle( math.atan2(dir[2], dir[1]) * 180 / math.pi)
	else
		self:setAngle( math.atan2(dir[2], -dir[1]) * 180 / math.pi)
	end
	
	self:setAngle(clamp(self:getAngle(), -self.angleRange, self.angleRange))
	
	dir = vec2.expd(self:getAngle())
	if e:getLeft() then dir[1] = -dir[1] end
	
	self:setPos{e:getPosX(), e:getPosY() - 1}
	self:setLeft(e:getLeft())
	self:setBlend(BLEND.ADD_ALPHA)
	
	self:setAnimGroup('sword')
	self:setScale{5,3}
	
	SwordLight:new(self)

--[[ instead of this junk, how about we just create an entity and collision-test it to all else?
-- oh... that'd require our non-aabb physics to work ... nevermind
--]]
	for o in allents() do
		if o and o:isa(TakesDamage) and o.health > 0 and o ~= e then
			local delta = vec2:new(o:getPos()) - self:getPos()
			local deltalen = delta:len()
			if delta:dot(dir) > .5*deltalen and deltalen < self.range then
				local tr,dist
				tr,dist = traceline(self:getPos(), delta, 0, true, false)
				if dist == 1 then
					o:setVel(o:getVel() + dir * (o:getMassInv() * self.force[1]))
					o:setVelX(o:getVelX() + o:getMassInv() * self.force[2] * dir[2])
					if not e:getLeft() then
						o:setVelY(o:getVelY() - o:getMassInv() * self.force[2] * dir[1])
					else
						o:setVelY(o:getVelY() + o:getMassInv() * self.force[2] * dir[1])
					end
					o:takeDamage(e, self.damage)
				end
			end
		end
	end
end

function Sword:think()
	Sword.super.think(self)
	
	local owner = self.holder
	if not owner then
		self:delete()
		return
	end
	
	local frac = (getTime() - self.createTime) / self.duration
	
	if frac < 0 or frac > 1 then
		self:delete()
		return
	end
	
	self:setPosX(self:getPosX() + (owner:getPosX() - self:getPosX()) * 0.0125)
	self:setPosY(self:getPosY() + (owner:getPosY() - 1 - self:getPosY()) * 0.0125)
	
	local c = 1 - frac*frac
	self:setColor{c,c,c,c}
end


weaponSword = WeaponClass:new(
	'Sword',
	nil,
	'Sword',
	0.3,
	function(e, aim)
		Sword:new(e, aim)		-- hmm, a local entity ... will this cause mismatches?
	end
)
